package mongoquery

import (
	"encoding/hex"
	"errors"
	"fmt"
	"github.com/globalsign/mgo"
	"github.com/globalsign/mgo/bson"
	"gitlab.okta-solutions.com/mashroom/backend/common/api/generic/query"
	"reflect"
	"regexp"
	"sort"
	"strconv"
	"strings"
	"time"
)

type EnumField struct {
	Interface interface{}
	Values    *map[string]int32
}

type enumFieldType struct {
	Type   reflect.Type
	Values *map[string]int32
}

type PipeOptions struct {
	Sort  bson.D
	Skip  int
	Limit int
}

var emptyValue = reflect.Value{}

func parseDateTime(timeString string) ([]time.Time, error) {
	dateFormats := []string{
		"Mon, 02 Jan 2006",
		"2006-01-02",
	}
	timeFormats := []string{
		time.RFC1123Z,    // "Mon, 02 Jan 2006 15:04:05 -0700" // RFC1123 with numeric zone
		time.RFC1123,     // "Mon, 02 Jan 2006 15:04:05 MST"
		time.RFC3339Nano, // "2006-01-02T15:04:05.999999999Z07:00"
		time.RFC3339,     // "2006-01-02T15:04:05Z07:00"
		"2006-01-02 03:04:05 PM -0700",
	}
	switch timeString {
	case "today":
		now := time.Now()
		year, month, day := now.Date()
		start := time.Date(year, month, day, 0, 0, 0, 0, now.Location())
		end := start.AddDate(0, 0, 1).Add(-1 * time.Nanosecond)
		return []time.Time{start, end}, nil
	case "yesterday":
		now := time.Now()
		year, month, day := now.Date()
		start := time.Date(year, month, day, 0, 0, 0, 0, now.Location()).AddDate(0, 0, -1)
		end := start.AddDate(0, 0, 1).Add(-1 * time.Nanosecond)
		return []time.Time{start, end}, nil
	case "this-week":
		now := time.Now()
		weekday := int(time.Now().Weekday())
		year, month, day := now.Date()
		start := time.Date(year, month, day, 0, 0, 0, 0, now.Location()).AddDate(0, 0, -weekday)
		end := start.AddDate(0, 0, 7).Add(-1 * time.Nanosecond)
		return []time.Time{start, end}, nil
	case "last-week":
		now := time.Now()
		weekday := int(time.Now().Weekday())
		year, month, day := now.Date()
		start := time.Date(year, month, day, 0, 0, 0, 0, now.Location()).AddDate(0, 0, -7-weekday)
		end := start.AddDate(0, 0, 7).Add(-1 * time.Nanosecond)
		return []time.Time{start, end}, nil
	}
	for _, dateFormat := range dateFormats {
		res, err := time.Parse(dateFormat, timeString)
		if err == nil {
			year, month, day := res.Date()
			start := time.Date(year, month, day, 0, 0, 0, 0, res.Location())
			end := start.AddDate(0, 0, 1).Add(-1 * time.Nanosecond)
			return []time.Time{start, end}, nil
		}
	}
	for _, timeFormat := range timeFormats {
		res, err := time.Parse(timeFormat, timeString)
		if err == nil {
			return []time.Time{res}, nil
		}
	}
	return nil, errors.New("error parsing date/time value")
}

func getFieldByTag(v interface{}, tagName, tagValue string) (fieldValue reflect.Value, fieldTagValue string) {
	if tagValue == "" {
		return emptyValue, ""
	}
	defer func() {
		if r := recover(); r != nil {
			fieldValue = emptyValue
		}
	}()

	var val reflect.Value
	val, k := v.(reflect.Value)
	if !k {
		val = reflect.ValueOf(v)
	}
	// if its a pointer, resolve its value
	if val.Kind() == reflect.Ptr {
		val = reflect.Indirect(val)
	}
	if val.Kind() == reflect.Struct {
		structType := val.Type()
		for i := 0; i < val.NumField(); i++ {
			fieldValue = val.Field(i)
			tags := strings.Split(structType.Field(i).Tag.Get(tagName), ",")
			if len(tags) > 0 {
				tag := strings.TrimSpace(tags[0])
				if (tag == tagValue) || ((tag == "_id") && (tagValue == "id")) {
					return fieldValue, tag
				}
			}
		}
	} else {
		return val, tagValue
	}
	return emptyValue, tagValue
}

func findFieldInner(item interface{}, fieldValue reflect.Value, fieldName *[]string, level int) (reflect.Value, string) {
	var field reflect.Value
	fName := (*fieldName)[level]
	if item != nil {
		field, fName = getFieldByTag(item, "bson", (*fieldName)[level])
	} else {
		field, fName = getFieldByTag(fieldValue, "bson", (*fieldName)[level])
	}
	(*fieldName)[level] = fName
	if (field != emptyValue) && (len(*fieldName) > level+1) {
		fName = (*fieldName)[level+1]
		field, fName = findFieldInner(nil, field, fieldName, level+1)
		(*fieldName)[level+1] = fName
	}
	return field, fName
}

func findField(item interface{}, fieldName string) (reflect.Value, string) {
	fName := strings.Split(fieldName, ".")
	fValue, _ := findFieldInner(item, reflect.Value{}, &fName, 0)
	return fValue, strings.Join(fName, ".")
}

func BuildFilter(model interface{}, filters []*query.QueryParamsFilter, enums []EnumField, override map[string]bson.M) (bson.M, error) {
	var res bson.M
	var filt []bson.M
	var enumTypes []enumFieldType
	for _, et := range enums {
		if et.Interface != nil {
			eType := reflect.TypeOf(et.Interface)
			if eType != nil {
				enumTypes = append(enumTypes, enumFieldType{Type: eType, Values: et.Values})
			}
		}
	}
	for _, f := range filters {
		fNamesOr := strings.Split(f.GetField(), ",")
		var filtOr []bson.M
		for _, ff := range fNamesOr {
			ff = strings.TrimSpace(ff)
			field, fieldName := findField(model, ff)
			if field == emptyValue {
				return nil, errors.New(fmt.Sprintf("unknown field \"%s\"", ff))
			}
			if override != nil {
				if _, ok := override[fieldName]; ok {
					continue
				}
			}
			v, err := generateMongoFilter(fieldName, field, f.GetValue(), f.GetOperator(), enumTypes)
			if err != nil {
				return nil, errors.New(fmt.Sprintf("%s for field \"%s\"", err, ff))
			}
			filtOr = append(filtOr, v)
		}
		switch len(filtOr) {
		case 0:
			return nil, errors.New(fmt.Sprintf("empty field name \"%s\"?", f.GetField()))
		case 1:
			filt = append(filt, filtOr[0])
		default:
			filt = append(filt, bson.M{"$or": filtOr})
		}
	}
	if override != nil {
		for _, over := range override {
			filt = append(filt, over)
		}
	}
	if len(filt) > 0 {
		res = bson.M{"$and": filt}
	}
	return res, nil
}

func BuildOptions(mQuery *mgo.Query, sorts []*query.QueryParamsSort, pagination *query.QueryParamsPagination) *PipeOptions {
	var sortOpt []string
	pipeOpt := PipeOptions{}
	for _, s := range sorts {
		if v := s.GetOrder(); (v == -1) || (v == 1) {
			so := "+"
			if v == -1 {
				so = "-"
			}
			so = so + s.GetField()
			sortOpt = append(sortOpt, so)
			pipeOpt.Sort = append(pipeOpt.Sort, bson.DocElem{Name: s.GetField(), Value: v})
		}
	}
	if len(sortOpt) > 0 {
		if mQuery != nil {
			mQuery.Sort(sortOpt...)
		}
	}
	if (pagination != nil) && (pagination.GetPerPage() > 0) {
		limit := int(pagination.GetPerPage())
		offset := (int(pagination.GetPage()) - 1) * limit
		if limit > 0 {
			if mQuery != nil {
				mQuery.Limit(limit)
			}
			pipeOpt.Limit = limit
		}
		if offset > 0 {
			if mQuery != nil {
				mQuery.Skip(offset)
			}
			pipeOpt.Skip = offset
		}
	}
	return &pipeOpt
}

func generateMongoFilter(fieldName string, fieldValue reflect.Value, values []string, op query.QueryFilterOperator, enumTypes []enumFieldType) (bson.M, error) {
	var first, second interface{}
	var all []interface{}
	if len(values) < 1 {
		return nil, nil
	}

	fieldType := fieldValue.Type()
	oId := reflect.TypeOf(bson.ObjectId(""))
	oTime := reflect.TypeOf(time.Time{})
	switch fieldType {
	case oId:
		for _, v := range values {
			if vv, err := hex.DecodeString(v); (err == nil) && (len(vv) == 12) {
				all = append(all, bson.ObjectId(vv))
			} else {
				return nil, errors.New("invalid id value")
			}
		}
		if op == query.QueryFilterOperator_UNKNOWN_OPERATOR {
			if len(all) > 1 {
				op = query.QueryFilterOperator_ONEOF
			} else {
				op = query.QueryFilterOperator_EQUALS
			}
		}
		availableOps := map[query.QueryFilterOperator]bool{
			query.QueryFilterOperator_ONEOF:  true,
			query.QueryFilterOperator_EQUALS: true,
			query.QueryFilterOperator_ANY:    true,
			query.QueryFilterOperator_ALL:    true,
		}
		if _, ok := availableOps[op]; !ok {
			return nil, errors.New("invalid operator for id")
		}
	case oTime:
		if op == query.QueryFilterOperator_UNKNOWN_OPERATOR {
			op = query.QueryFilterOperator_EQUALS
		}
		opIn := op
		for _, v := range values {
			if vv, err := parseDateTime(v); err == nil {
				for _, timeValue := range vv {
					all = append(all, timeValue)
				}
				if op == query.QueryFilterOperator_EQUALS && len(vv) > 1 {
					op = query.QueryFilterOperator_RANGE
					break
				}
			} else {
				return nil, err
			}
		}
		if opIn == query.QueryFilterOperator_RANGE && len(all) > 2 {
			sort.Slice(all, func(i, j int) bool {
				timeDiff := all[i].(time.Time).Sub(all[j].(time.Time))
				return timeDiff.Nanoseconds() < 0
			})
			all = []interface{}{all[0], all[len(all)-1]}
		}
	default:
		for _, eType := range enumTypes {
			if eType.Type == fieldType {
				for _, v := range values {
					if vv, ok := (*eType.Values)[v]; ok {
						all = append(all, vv)
					} else {
						return nil, errors.New(fmt.Sprintf("invalid enum value \"%s\"", v))
					}
				}
				break
			}
		}
	}

	if len(all) == 0 {
		switch fieldValue.Kind() {
		case reflect.String:
			for _, v := range values {
				all = append(all, v)
			}
			if op == query.QueryFilterOperator_UNKNOWN_OPERATOR {
				if len(all) > 1 {
					op = query.QueryFilterOperator_ONEOF
				} else {
					op = query.QueryFilterOperator_EQUALS
				}
			}
		case reflect.Bool:
			for _, v := range values {
				truth := map[string]bool{"1": true, "on": true, "true": true, "0": false, "off": false, "false": false}
				vv := false
				if v, ok := truth[v]; ok {
					vv = v
				} else {
					return nil, errors.New("invalid boolean value")
				}
				all = append(all, vv)
			}
			if op == query.QueryFilterOperator_UNKNOWN_OPERATOR {
				op = query.QueryFilterOperator_EQUALS
			}
		case reflect.Int, reflect.Int8, reflect.Int16, reflect.Int32, reflect.Int64:
			for _, v := range values {
				if vv, err := strconv.ParseInt(v, 10, 64); err == nil {
					all = append(all, vv)
				} else {
					return nil, errors.New("invalid numeric value")
				}
			}
			if op == query.QueryFilterOperator_UNKNOWN_OPERATOR {
				if len(all) > 1 {
					op = query.QueryFilterOperator_ONEOF
				} else {
					op = query.QueryFilterOperator_EQUALS
				}
			}
		case reflect.Uint, reflect.Uint8, reflect.Uint16, reflect.Uint32, reflect.Uint64:
			for _, v := range values {
				if vv, err := strconv.ParseUint(v, 10, 64); err == nil {
					all = append(all, vv)
				} else {
					return nil, errors.New("invalid numeric value")
				}
			}
			if op == query.QueryFilterOperator_UNKNOWN_OPERATOR {
				if len(all) > 1 {
					op = query.QueryFilterOperator_ONEOF
				} else {
					op = query.QueryFilterOperator_EQUALS
				}
			}
		case reflect.Float32, reflect.Float64:
			for _, v := range values {
				if vv, err := strconv.ParseFloat(v, 64); err == nil {
					all = append(all, vv)
				} else {
					return nil, errors.New("invalid float value")
				}
			}
			if op == query.QueryFilterOperator_UNKNOWN_OPERATOR {
				if len(all) > 1 {
					op = query.QueryFilterOperator_ONEOF
				} else {
					op = query.QueryFilterOperator_EQUALS
				}
			}
		default:
			return nil, errors.New(fmt.Sprintf("type \"%s\" not supported yet", fieldValue.Type().String()))
		}
	}
	if len(all) == 0 {
		return nil, nil
	}
	first = all[0]
	if len(all) > 1 {
		second = all[1]
	}

	switch op {
	case query.QueryFilterOperator_EQUALS:
		return bson.M{fieldName: bson.M{"$eq": first}}, nil
	case query.QueryFilterOperator_CONTAINS:
		fs, ok := first.(string)
		if !ok {
			return nil, errors.New("invalid operator")
		}
		return bson.M{fieldName: bson.M{"$regex": regexp.QuoteMeta(fs), "$options": "i"}}, nil
	case query.QueryFilterOperator_CONTAINS_ANY:
		res := make([]bson.M, len(all))
		for _, s := range all {
			fs, ok := s.(string)
			if !ok {
				return nil, errors.New("invalid operator")
			}
			res = append(res, bson.M{"$regex": regexp.QuoteMeta(fs)})
		}
		return bson.M{fieldName: bson.M{"$or": res}}, nil
	case query.QueryFilterOperator_DASH:
		fs, ok := first.(string)
		if !ok {
			return nil, errors.New("invalid operator")
		}
		fs = strings.ReplaceAll(fs, "  ", " ")
		fs = strings.ReplaceAll(fs, "-", " ")
		sc := strings.ReplaceAll(fs, " ", "-")
		return bson.M{fieldName: bson.M{"$in": []string{fs, sc}}}, nil
	case query.QueryFilterOperator_DASH_CONTAINS:
		fs, ok := first.(string)
		if !ok {
			return nil, errors.New("invalid operator")
		}
		fs = strings.ReplaceAll(fs, "-", " ")
		fs = strings.ReplaceAll(fs, "\t", " ")
		fSplit := make([]string, 0)
		for _, s := range strings.Split(fs, " ") {
			if len(s) > 0 {
				fSplit = append(fSplit, regexp.QuoteMeta(s))
			}
		}
		fs = strings.Join(fSplit, ".*[ -].*")
		return bson.M{fieldName: bson.M{"$regex": fs, "$options": "i"}}, nil
	case query.QueryFilterOperator_DASH_ANY:
		fs, ok := first.(string)
		if !ok {
			return nil, errors.New("invalid operator")
		}
		fs = strings.ReplaceAll(fs, "-", " ")
		fs = strings.ReplaceAll(fs, "\t", " ")
		fSplit := make([]bson.M, 0)
		for _, s := range strings.Split(fs, " ") {
			if len(s) > 0 {
				fSplit = append(fSplit, bson.M{fieldName: bson.M{"$regex": regexp.QuoteMeta(s)}})
			}
		}
		return bson.M{"$or": fSplit}, nil
	case query.QueryFilterOperator_ONEOF:
		return bson.M{fieldName: bson.M{"$in": all}}, nil
	case query.QueryFilterOperator_MORE:
		return bson.M{fieldName: bson.M{"$gte": first}}, nil
	case query.QueryFilterOperator_LESS:
		return bson.M{fieldName: bson.M{"$lte": first}}, nil
	case query.QueryFilterOperator_RANGE:
		if second == nil {
			return nil, errors.New("two values needed")
		}
		return bson.M{fieldName: bson.M{
			"$gte": first,
			"$lte": second,
		}}, nil
	case query.QueryFilterOperator_ANY:
		return bson.M{fieldName: bson.M{"$in": all}}, nil
	case query.QueryFilterOperator_ALL:
		return bson.M{fieldName: bson.M{"$all": all}}, nil
	case query.QueryFilterOperator_STARTWITH:
		fs, ok := first.(string)
		if !ok {
			return nil, nil
		}
		return bson.M{fieldName: bson.M{"$regex": "^" + regexp.QuoteMeta(fs), "$options": "i"}}, nil
	}
	return nil, errors.New("unknown operator")
}
