package impl

import (
	"github.com/globalsign/mgo"
	"github.com/globalsign/mgo/bson"
	"github.com/golang/protobuf/ptypes/timestamp"
	"gitlab.okta-solutions.com/mashroom/backend/cardpayment/api"
	"gitlab.okta-solutions.com/mashroom/backend/cardpayment/impl/mongoquery"
	q "gitlab.okta-solutions.com/mashroom/backend/common/api/generic/query"
	"gitlab.okta-solutions.com/mashroom/backend/common/generic/env"
	"gitlab.okta-solutions.com/mashroom/backend/common/generic/log"
	"gitlab.okta-solutions.com/mashroom/backend/extras/api"
	"strings"
	"time"
)

const (
	MongoPropertiesCollection = "charges"
)

var (
	MongoUrl      = env.ResolveString("MONGO_URL", "mongodb:27017")
	MongoUsername = env.ResolveString("MONGO_USERNAME", "cardpayment-data")
	MongoPassword = env.ResolveString("MONGO_PASSWORD", "cardpayment-data")
	MongoAuthDb   = env.ResolveString("MONGO_AUTH", "admin")
	MongoDb       = env.ResolveString("MONGO_DATABASE", "cardpayment-data")
)

type ConfirmModel struct {
	Admin       string    `bson:"admin"`
	Confirmed   bool      `bson:"confirmed"`
	ConfirmedAt time.Time `bson:"confirmed_at"`
}

func (model *ConfirmModel) Update(value bool, admin string) bool {
	res := false
	if value && !model.Confirmed {
		model.Confirmed = true
		model.ConfirmedAt = time.Now().UTC()
		model.Admin = admin
		res = true
	} else if !value && model.Confirmed {
		model.Confirmed = false
		model.ConfirmedAt = time.Unix(0, 0).UTC()
		model.Admin = ""
		res = true
	}
	return res
}

type ExtraServiceModel struct {
	ServiceId   string               `bson:"_id"`
	Id          string               `bson:"item_id"`
	Type        extras.ServiceType   `bson:"type"`
	Code        string               `bson:"code"`
	Status      extras.ServiceStatus `bson:"status"`
	Price       float64              `bson:"price"`
	Description string               `bson:"description"`
	Primary     bool                 `bson:"primary"`
	Created     time.Time            `bson:"created"`
	Updated     time.Time            `bson:"updated"`
	Confirm     *ConfirmModel        `bson:"confirm"`
	Role        extras.ServiceRole   `bson:"role"`
}

type CardpaymentModel struct {
	Id            string                    `bson:"_id,omitempty"`
	Type          cardpayment.PaymentType   `bson:"type"`
	ClientSecret  string                    `bson:"client_secret,omitempty"`
	UserId        string                    `bson:"user_id,omitempty"`
	UserEmail     string                    `bson:"user_email,omitempty"`
	Items         []*ExtraServiceModel      `bson:"items,omitempty"`
	Status        cardpayment.PaymentStatus `bson:"status,omitempty"`
	Property      string                    `bson:"property,omitempty"`
	Total         int64                     `bson:"total,omitempty"`
	NotifyId      string                    `bson:"notify_id"`
	NotifyAddress string                    `bson:"notify_address"`
	CreatedAt     time.Time                 `bson:"created_at,omitempty"`
	UpdatedAt     time.Time                 `bson:"updated_at,omitempty"`
	InvoiceNumber uint64                    `bson:"invoice_number"`
}

type CardpaymentItemModel struct {
	ChargeId      string                    `bson:"charge_id,omitempty"`
	ChargeType    cardpayment.PaymentType   `bson:"charge_type,omitempty"`
	ClientSecret  string                    `bson:"client_secret,omitempty"`
	UserId        string                    `bson:"user_id,omitempty"`
	UserEmail     string                    `bson:"user_email,omitempty"`
	ChargeStatus  cardpayment.PaymentStatus `bson:"charge_status,omitempty"`
	Property      string                    `bson:"property,omitempty"`
	Total         int64                     `bson:"total,omitempty"`
	NotifyId      string                    `bson:"notify_id"`
	NotifyAddress string                    `bson:"notify_address"`
	CreatedAt     time.Time                 `bson:"created_at,omitempty"`
	UpdatedAt     time.Time                 `bson:"updated_at,omitempty"`
	InvoiceNumber uint64                    `bson:"invoice_number"`

	ServiceId   string               `bson:"service_id"`
	Id          string               `bson:"item_id"`
	Type        extras.ServiceType   `bson:"type"`
	Code        string               `bson:"code"`
	Status      extras.ServiceStatus `bson:"status"`
	Price       float64              `bson:"price"`
	Description string               `bson:"description"`
	Primary     bool                 `bson:"primary"`
	Created     time.Time            `bson:"created"`
	Updated     time.Time            `bson:"updated"`
	Confirm     *ConfirmModel        `bson:"confirm"`
}

type ItemsCount struct {
	Count int64 `bson:"count"`
}

func ToUnixTime(t time.Time) *timestamp.Timestamp {
	return &timestamp.Timestamp{
		Seconds: t.Unix(),
		Nanos:   int32(t.Nanosecond()),
	}
}

func UpdateServiceModel(model *ExtraServiceModel, s *extras.Service) {
	model.Type = s.GetTypeValue()
	model.Code = s.GetCodeValue()
	model.Status = s.GetStatusValue()
	model.Price = s.GetPriceValue()
	model.Description = s.GetDescriptionValue()
	model.Primary = s.GetPrimaryValue()
	model.Role = s.GetRoleValue()
}

func RenderServiceModel(model *ExtraServiceModel) *extras.Service {
	if model.Confirm == nil {
		model.Confirm = &ConfirmModel{}
	}
	var netPrice float64 = 0.0
	var taxAmount float64 = 0.0

	//VAT on Income 20%
	if model.Role == extras.ServiceRole_LANDLORD_LISTING {
		netPrice = (model.Price / 6) * 5
		taxAmount = model.Price / 6

		netPrice = Round(netPrice, 2)
		taxAmount = Round(taxAmount, 2)
	}

	return &extras.Service{
		Id:          &extras.Service_IdValue{IdValue: model.Id},
		ItemId:      &extras.Service_ItemIdValue{ItemIdValue: model.Id},
		Type:        &extras.Service_TypeValue{TypeValue: model.Type},
		Code:        &extras.Service_CodeValue{CodeValue: model.Code},
		Status:      &extras.Service_StatusValue{StatusValue: model.Status},
		Price:       &extras.Service_PriceValue{PriceValue: model.Price},
		Description: &extras.Service_DescriptionValue{DescriptionValue: model.Description},
		Primary:     &extras.Service_PrimaryValue{PrimaryValue: model.Primary},
		Confirm: &extras.Service_ConfirmValue{ConfirmValue: &extras.ServiceConfirm{
			Admin:       model.Confirm.Admin,
			Confirmed:   model.Confirm.Confirmed,
			ConfirmedAt: ToUnixTime(model.Confirm.ConfirmedAt),
		}},
		Role:      &extras.Service_RoleValue{RoleValue: model.Role},
		NetPrice:  &extras.Service_NetPriceValue{NetPriceValue: netPrice},
		TaxAmount: &extras.Service_TaxAmountValue{TaxAmountValue: taxAmount},
	}
}

type Repo struct {
	db      *mgo.Database
	charges *mgo.Collection
}

func (repo *Repo) FixItemIds() error {
	query := bson.M{"$or": []bson.M{{"items.item_id": bson.M{"$exists": false}}, {"items.item_id": ""}, {"items.item_id": bson.M{"$regex": "ObjectId"}}}}
	var result []*CardpaymentModel
	err := repo.charges.Find(query).All(&result)
	if err != nil {
		return err
	}
	for _, charge := range result {
		fixed := false
		for _, item := range charge.Items {
			if item.Id == "" || strings.HasPrefix(item.Id, "ObjectId") {
				item.Id = bson.NewObjectId().Hex()
				log.Info("fixed charge item ", charge.Id, " with new item_id ", item.Id)
				fixed = true
			}
		}
		if fixed {
			if err = repo.Update(charge); err != nil {
				log.Error("error updating charge item", err)
			}
		}
	}
	return nil
}

func (repo *Repo) Create(model *CardpaymentModel) error {
	now := time.Now()
	model.CreatedAt = now
	model.UpdatedAt = now
	if model.Type == cardpayment.PaymentType_UNKNOWN_PAYMENT_TYPE {
		model.Type = cardpayment.PaymentType_CARD
	}
	return repo.charges.Insert(model)
}

func (repo *Repo) CreateEnsureId(model *CardpaymentModel, idPrefix string) error {
	if model.Id == "" {
		model.Id = idPrefix + bson.NewObjectId().Hex()
	}
	return repo.Create(model)
}

func (repo *Repo) Update(model *CardpaymentModel) error {
	model.UpdatedAt = time.Now()
	if model.Type == cardpayment.PaymentType_UNKNOWN_PAYMENT_TYPE {
		model.Type = cardpayment.PaymentType_CARD
	}
	return repo.charges.UpdateId(model.Id, model)
}

func (repo *Repo) GetByUser(user string) ([]*CardpaymentModel, error) {
	query := bson.M{"user_id": user}
	var result []*CardpaymentModel
	err := repo.charges.Find(query).All(&result)
	return result, err
}

func (repo *Repo) GetByUserProperty(user, property string) ([]*CardpaymentModel, error) {
	query := bson.M{"user_id": user, "property": property}
	var result []*CardpaymentModel
	err := repo.charges.Find(query).All(&result)
	return result, err
}

func (repo *Repo) GetByProperty(property string) ([]*CardpaymentModel, error) {
	query := bson.M{"property": property}
	var result []*CardpaymentModel
	err := repo.charges.Find(query).All(&result)
	return result, err
}

func (repo *Repo) ConfirmChargeItem(chargeId, itemId, admin string, confirm bool) (*CardpaymentItemModel, error) {
	return repo.doGetConfirmChargeItem(chargeId, itemId, admin, confirm)
}

func (repo *Repo) GetChargeItem(chargeId, itemId string) (*CardpaymentItemModel, error) {
	return repo.doGetConfirmChargeItem(chargeId, itemId, "", false)
}

func (repo *Repo) doGetConfirmChargeItem(chargeId, itemId, admin string, confirm bool) (*CardpaymentItemModel, error) {
	query := bson.M{"$and": []bson.M{{"_id": chargeId}, {"items.item_id": itemId}}}
	chargesQuery := repo.charges.Find(query)
	var rcpts []*CardpaymentModel
	err := chargesQuery.All(&rcpts)
	if err != nil {
		return nil, err
	}
	res := &CardpaymentItemModel{}
	for _, r := range rcpts {
		update := false
		for _, item := range r.Items {
			if item.Id == itemId {
				if item.Confirm == nil {
					item.Confirm = &ConfirmModel{}
				}
				if admin != "" {
					update = item.Confirm.Update(confirm, admin)
				}
				if r.Type == cardpayment.PaymentType_UNKNOWN_PAYMENT_TYPE {
					r.Type = cardpayment.PaymentType_CARD
				}
				res = &CardpaymentItemModel{
					ChargeId:      r.Id,
					ChargeType:    r.Type,
					ClientSecret:  r.ClientSecret,
					UserId:        r.UserId,
					UserEmail:     r.UserEmail,
					ChargeStatus:  r.Status,
					Property:      r.Property,
					Total:         r.Total,
					NotifyId:      r.NotifyId,
					NotifyAddress: r.NotifyAddress,
					CreatedAt:     r.CreatedAt,
					UpdatedAt:     r.UpdatedAt,
					InvoiceNumber: r.InvoiceNumber,

					ServiceId:   item.ServiceId,
					Id:          item.Id,
					Type:        item.Type,
					Code:        item.Code,
					Status:      item.Status,
					Price:       item.Price,
					Description: item.Description,
					Primary:     item.Primary,
					Created:     item.Created,
					Updated:     item.Updated,
					Confirm:     item.Confirm,
				}
				break
			}
		}
		if update {
			if err = repo.Update(r); err != nil {
				// TODO: log? return err?
			}
		}
	}
	return res, nil
}

func (repo *Repo) FilterCharges(params *q.QueryParams, admin bool) ([]*CardpaymentModel, int32, error) {
	enums := []mongoquery.EnumField{
		{
			Interface: cardpayment.Currency(0),
			Values:    &cardpayment.Currency_value,
		},
		{
			Interface: cardpayment.PaymentType(0),
			Values:    &cardpayment.PaymentType_value,
		},
		{
			Interface: cardpayment.PaymentStatus(0),
			Values:    &cardpayment.PaymentStatus_value,
		},
		{
			Interface: extras.ServiceType(0),
			Values:    &extras.ServiceType_value,
		},
		{
			Interface: extras.ServiceStatus(0),
			Values:    &extras.ServiceStatus_value,
		},
		{
			Interface: extras.ServiceRole(0),
			Values:    &extras.ServiceRole_value,
		},
	}
	query, err := mongoquery.BuildFilter(CardpaymentModel{}, params.GetFilter(), enums, nil)
	if err != nil {
		return nil, 0, err
	}
	chargesQuery := repo.charges.Find(query)
	mongoquery.BuildOptions(chargesQuery, params.GetSort(), params.GetPagination())
	var result []*CardpaymentModel
	err = chargesQuery.All(&result)
	if err != nil {
		return nil, 0, err
	}
	total, err := repo.charges.Find(query).Count()
	return result, int32(total), err
}

func (repo *Repo) FilterChargeItems(params *q.QueryParams, admin bool) ([]*CardpaymentItemModel, int32, int32, error) {
	enums := []mongoquery.EnumField{
		{
			Interface: cardpayment.Currency(0),
			Values:    &cardpayment.Currency_value,
		},
		{
			Interface: cardpayment.PaymentType(0),
			Values:    &cardpayment.PaymentType_value,
		},
		{
			Interface: cardpayment.PaymentStatus(0),
			Values:    &cardpayment.PaymentStatus_value,
		},
		{
			Interface: extras.ServiceType(0),
			Values:    &extras.ServiceType_value,
		},
		{
			Interface: extras.ServiceStatus(0),
			Values:    &extras.ServiceStatus_value,
		},
		{
			Interface: extras.ServiceRole(0),
			Values:    &extras.ServiceRole_value,
		},
	}
	filter := params.GetFilter()
	for i, _ := range filter {
		if filter[i].Field == "id" {
			filter[i].Field = "item_id"
		}
	}
	sorting := params.GetSort()
	for i, _ := range sorting {
		if sorting[i].Field == "id" {
			sorting[i].Field = "item_id"
		}
	}
	query, err := mongoquery.BuildFilter(CardpaymentItemModel{}, filter, enums, nil)
	if err != nil {
		return nil, 0, 0, err
	}
	opt := mongoquery.BuildOptions(nil, sorting, params.GetPagination())

	project := bson.M{
		"$project": bson.M{
			"charge_id":      "$_id",
			"charge_type":    "$type",
			"client_secret":  "$client_secret",
			"charge_status":  "$status",
			"user_id":        "$user_id",
			"user_email":     "$user_email",
			"property":       "$property",
			"total":          "$total",
			"notify_id":      "$notify_id",
			"notify_address": "$notify_address",
			"created_at":     "$created_at",
			"updated_at":     "$updated_at",
			"invoice_number": "$invoice_number",
			"service_id":     "$items._id",
			"item_id":        "$items.item_id",
			"type":           "$items.type",
			"code":           "$items.code",
			"status":         "$items.status",
			"price":          "$items.price",
			"description":    "$items.description",
			"primary":        "$items.primary",
			"created":        "$items.created",
			"updated":        "$items.updated",
			"confirm":        "$items.confirm",
		},
	}
	pipeline := []bson.M{
		{"$unwind": "$items"},
		project,
	}
	if len(query) > 0 {
		pipeline = append(pipeline, bson.M{"$match": query})
	}
	if len(opt.Sort) > 0 {
		pipeline = append(pipeline, bson.M{"$sort": opt.Sort})
	}
	if opt.Limit > 0 {
		pipeline = append(pipeline, bson.M{"$limit": opt.Skip + opt.Limit})
	}
	if opt.Skip > 0 {
		pipeline = append(pipeline, bson.M{"$skip": opt.Skip})
	}
	chargesQuery := repo.charges.Pipe(pipeline)
	var result []*CardpaymentItemModel
	err = chargesQuery.All(&result)
	if err != nil {
		return nil, 0, 0, err
	}
	var nonConfirmed int32 = 0
	total := int32(len(result))
	pipeline = []bson.M{
		{"$unwind": "$items"},
		project,
	}
	if len(query) > 0 {
		pipeline = append(pipeline, bson.M{"$match": query})
	}
	pipeline = append(pipeline, bson.M{"$group": bson.M{"_id": 0, "count": bson.M{"$sum": 1}}})
	pipeline = append(pipeline, bson.M{"$project": bson.M{"_id": 0}})
	chargesCountQuery := repo.charges.Pipe(pipeline)
	var resCount []*ItemsCount
	err = chargesCountQuery.All(&resCount)
	if err != nil {
		return nil, 0, 0, err
	}
	if len(resCount) > 0 {
		total = int32(resCount[0].Count)
	}
	//for _, v := range result {
	//	if !v.Confirm.Confirmed {
	//		nonConfirmed = nonConfirmed + 1
	//	}
	//}
	return result, total, nonConfirmed, err
}

func (repo *Repo) GetByID(id string) (*CardpaymentModel, error) {
	query := bson.M{"_id": id}
	var result = &CardpaymentModel{}
	err := repo.charges.Find(query).One(result)
	return result, err
}

func NewRepoFromEnv() (*Repo, error) {
	return NewRepo(&mgo.DialInfo{
		Addrs:    []string{MongoUrl},
		Username: MongoUsername,
		Password: MongoPassword,
		Source:   MongoAuthDb,
		Database: MongoDb,
	}, MongoPropertiesCollection)
}

func NewRepo(dialInfo *mgo.DialInfo, collectionName string) (*Repo, error) {
	log.WithFields(log.Fields{"dialInfo": dialInfo}).Info("Connecting to mongo")
	session, err := mgo.DialWithInfo(dialInfo)
	if err != nil {
		return nil, err
	}
	db := session.DB(MongoDb)
	log.Info("Mongo connected.")

	collection := db.C(collectionName)

	return &Repo{
		db:      db,
		charges: collection,
	}, nil
}
