package impl

import (
	"bytes"
	"context"
	"crypto/sha256"
	"encoding/hex"
	"encoding/json"
	"errors"
	"fmt"
	"github.com/globalsign/mgo/bson"
	"github.com/golang/protobuf/ptypes"
	"github.com/stripe/stripe-go"
	_ "github.com/stripe/stripe-go/customer"
	"github.com/stripe/stripe-go/paymentintent"
	"github.com/zeebe-io/zeebe/clients/go/zbc"
	account "gitlab.okta-solutions.com/mashroom/backend/account/api"
	"gitlab.okta-solutions.com/mashroom/backend/cardpayment/api"
	"gitlab.okta-solutions.com/mashroom/backend/cardpayment/version"
	"gitlab.okta-solutions.com/mashroom/backend/common/api/health"
	"gitlab.okta-solutions.com/mashroom/backend/common/generic/bind"
	"gitlab.okta-solutions.com/mashroom/backend/common/generic/env"
	"gitlab.okta-solutions.com/mashroom/backend/common/generic/grpc"
	"gitlab.okta-solutions.com/mashroom/backend/common/generic/log"
	contract "gitlab.okta-solutions.com/mashroom/backend/contract/api"
	document "gitlab.okta-solutions.com/mashroom/backend/document/api"
	"gitlab.okta-solutions.com/mashroom/backend/extras/api"
	messaging "gitlab.okta-solutions.com/mashroom/backend/messaging/api"
	"gitlab.okta-solutions.com/mashroom/backend/notification/api"
	"gitlab.okta-solutions.com/mashroom/backend/property/api"
	"io"
	"math"
	"net"
	"strconv"
	"strings"
	"time"
)

var ExtrasServiceUrl = env.ResolveString("EXTRAS_SERVICE_URL", "mashroom-extras:10000")
var ExtrasService = extras.NewExtrasServiceClient(grpc.Dial(ExtrasServiceUrl, "Extras"))

var NotificationServiceUrl = env.ResolveString("NOTIFICATION_SERVICE_URL", "mashroom-notification:10000")
var NotificationService = notification.NewNotificationServiceClient(grpc.Dial(NotificationServiceUrl, "Notification"))

var PropertyServiceUrl = env.ResolveString("PROPERTY_SERVICE_URL", "mashroom-property:10000")
var PropertyService = property.NewPropertyServiceClient(grpc.Dial(PropertyServiceUrl, "Property"))

var AccountServiceUrl = env.ResolveString("ACCOUNT_SERVICE_URL", "mashroom-account:10000")
var AccountService = account.NewAccountServiceClient(grpc.Dial(AccountServiceUrl, "Account"))

var DocumentServiceUrl = env.ResolveString("DOCUMENT_SERVICE_URL", "mashroom-document:10000")
var DocumentService = document.NewDocumentServiceClient(grpc.Dial(DocumentServiceUrl, "Document"))

var ContractServiceUrl = env.ResolveString("CONTRACT_SERVICE_URL", "mashroom-contract:10000")
var ContractService = contract.NewContractServiceClient(grpc.Dial(ContractServiceUrl, "Contract"))

var stripeCurrency = stripe.CurrencyGBP

type Server interface {
	cardpayment.CardpaymentServiceServer
	Serve(addr string)
	Background()
}

type serverImpl struct {
	repo     *Repo
	ZBClient zbc.ZBClient
}

var ErrInvalidInit = errors.New("invalid init")
var ErrInvalidChunk = errors.New("invalid chunk")
var ErrInvalidHash = errors.New("invalid hash")
var ErrNotOwned = errors.New("not owned receipt")
var ErrCancelled = errors.New("receipt cancelled")

func init() {
	cur := strings.ToUpper(env.ResolveString("STRIPE_CURRENCY", "GBP"))
	switch cur {
	case "GBP":
		stripeCurrency = stripe.CurrencyGBP
	case "EUR":
		stripeCurrency = stripe.CurrencyEUR
	case "USD":
		stripeCurrency = stripe.CurrencyUSD
	}
}

func (server *serverImpl) IssueUpdate(m *CardpaymentModel) {
	_, _ = server.ZBClient.NewPublishMessageCommand().MessageName("payment-update").CorrelationKey(m.Property).Send()

	//_, _ = NotificationService.SendSms(context.Background(), &notification.SendSmsRequest{
	//	To:   "+79996449060",
	//	Body: "Test",
	//})

	if m.NotifyAddress != "" {
		notif := cardpayment.NewCardpaymentNotifyServiceClient(grpc.Dial(m.NotifyAddress, "NotifyCardpayment"))
		_, err := notif.CardpaymentNotify(context.TODO(), &cardpayment.CardpaymentNotifyRequest{
			NotifyId: m.NotifyId,
			Receipt:  RenderModel(m),
		})
		if err != nil {
			log.Error(err)
		}
	}
	_, _ = PropertyService.Update(context.Background(), &property.UpdateRequest{
		Property: &property.Property{
			Id: &property.Property_IdValue{IdValue: m.Property},
		},
	})
}

func (server *serverImpl) HandleCallback(stream cardpayment.CardpaymentService_HandleCallbackServer) error {
	ev, err := stream.Recv()
	if err != nil {
		return err
	}
	if ev.Init == nil {
		return ErrInvalidInit
	}

	remote := ev.Init.Sha256

	var data []byte
	for {
		ev, err := stream.Recv()
		if err != nil {
			if err == io.EOF || err == io.ErrUnexpectedEOF {
				break
			}
			return err
		}
		if ev.Chunk == nil {
			return ErrInvalidChunk
		}
		if len(ev.Chunk.Data) == 0 {
			break
		}

		data = append(data, ev.Chunk.Data...)
	}
	fmt.Println(string(data))

	h := sha256.New()
	actual := h.Sum(data)

	if !bytes.Equal(remote, actual) {
		fmt.Println("invalid hash", hex.EncodeToString(remote), "actual", hex.EncodeToString(actual))
		return ErrInvalidHash
	}

	resp, err := HandleCallbackImpl(data)
	if err != nil {
		fmt.Println("err callback", err)
		return err
	}

	m, err := server.repo.GetByID(resp.ID)
	if err != nil {
		return err
	}
	m.Status = resp.Status
	err = server.repo.Update(m)
	if err != nil {
		fmt.Println("err update", err)
		return err
	}

	server.IssueUpdate(m)
	_ = stream.SendAndClose(nil)
	return nil
}

func (server *serverImpl) FormReceipt(ctx context.Context, request *cardpayment.FormReceiptRequest) (*cardpayment.FormReceiptResponse, error) {
	var amount float64
	var extraitems []*extras.Service
	for _, i := range request.GetItems() {
		e, err := ExtrasService.Get(context.Background(), &extras.GetRequest{
			Id:         i,
			PropertyId: request.Property,
		})
		if err != nil {
			return nil, err
		}
		extraitems = append(extraitems, e.ExtraService)
		amount += e.ExtraService.GetPriceValue()
	}
	paymentType := request.GetType()
	if paymentType == cardpayment.PaymentType_UNKNOWN_PAYMENT_TYPE {
		paymentType = cardpayment.PaymentType_CARD
	}
	model := &CardpaymentModel{
		//Id:            intent.ID,
		//ClientSecret:  intent.ClientSecret,
		//Items:         items,
		Type:          paymentType,
		UserId:        request.GetUserId(),
		UserEmail:     request.GetUserEmail(),
		Status:        cardpayment.PaymentStatus_PROCESSING,
		Property:      request.GetProperty(),
		NotifyId:      request.GetNotifyId(),
		NotifyAddress: request.GetNotifyAddress(),
	}
	fallbackPrefix := ""
	if amount == 0 {
		fallbackPrefix = "null_"
		model.Status = cardpayment.PaymentStatus_SUCCEEDED
	} else if paymentType == cardpayment.PaymentType_BANK_ACCOUNT {
		fallbackPrefix = "bank_"
		model.Status = cardpayment.PaymentStatus_SUCCEEDED
	} else {
		bs, err := json.Marshal(request.GetItems())
		if err != nil {
			return nil, err
		}
		if amount < 0.5 {
			return nil, errors.New(fmt.Sprintf("too small amount to charge (%.2f < 0.50)", amount))
		}
		intent, err := paymentintent.New(&stripe.PaymentIntentParams{
			Params: stripe.Params{
				Metadata: map[string]string{
					"propertyId": request.GetProperty(),
					"userid":     request.GetUserId(),
					"items":      string(bs),
				},
			},
			Amount:             stripe.Int64(int64(math.Ceil(amount * 100))),
			Currency:           stripe.String(string(stripeCurrency)),
			CaptureMethod:      stripe.String("automatic"),
			PaymentMethodTypes: []*string{stripe.String("card")},
			ReceiptEmail:       stripe.String(request.GetUserEmail()),
		})
		if err != nil {
			return nil, err
		}
		model.Id = intent.ID
		model.ClientSecret = intent.ClientSecret
	}
	var items []*ExtraServiceModel
	for _, e := range extraitems {
		m := &ExtraServiceModel{
			ServiceId: e.GetIdValue(),
			Id:        bson.NewObjectId().Hex(),
		}
		UpdateServiceModel(m, e)
		items = append(items, m)
	}
	model.Items = items
	err := server.repo.CreateEnsureId(model, fallbackPrefix)
	if err != nil {
		return nil, err
	}

	return &cardpayment.FormReceiptResponse{
		Id:           model.Id,
		ClientSecret: model.ClientSecret,
	}, nil
}

func (server *serverImpl) CancelReceipt(ctx context.Context, request *cardpayment.CancelReceiptRequest) (*cardpayment.CancelReceiptResponse, error) {
	m, err := server.repo.GetByID(request.GetId())
	if err != nil {
		return nil, err
	}
	if m.UserId != request.GetUserId() {
		return nil, ErrNotOwned
	}
	m.Status = cardpayment.PaymentStatus_CANCELLED
	err = server.repo.Update(m)
	if err != nil {
		return nil, err
	}
	_, err = paymentintent.Cancel(m.Id, &stripe.PaymentIntentCancelParams{
		CancellationReason: stripe.String(string(stripe.PaymentIntentCancellationReasonRequestedByCustomer)),
	})
	if err != nil {
		return nil, err
	}
	return &cardpayment.CancelReceiptResponse{
		Success: true,
	}, nil
}

func (server *serverImpl) CheckReceipt(ctx context.Context, request *cardpayment.CheckReceiptRequest) (*cardpayment.CheckReceiptResponse, error) {
	m, err := server.repo.GetByID(request.GetId())
	if err != nil {
		return nil, err
	}
	if m.UserId != request.GetUserId() {
		return nil, ErrNotOwned
	}
	intent, err := paymentintent.Get(request.GetId(), &stripe.PaymentIntentParams{})
	if err != nil {
		return nil, err
	}

	if intent.Status == stripe.PaymentIntentStatusSucceeded && m.Status != cardpayment.PaymentStatus_SUCCEEDED {
		m.Status = cardpayment.PaymentStatus_SUCCEEDED
		err = server.repo.Update(m)
		if err != nil {
			return nil, err
		}
		server.IssueUpdate(m)
	}
	return &cardpayment.CheckReceiptResponse{
		Receipt: RenderModel(m),
	}, nil
}

func RenderModel(model *CardpaymentModel) *cardpayment.Receipt {
	var items []*extras.Service
	for _, s := range model.Items {
		items = append(items, RenderServiceModel(s))
	}
	created, _ := ptypes.TimestampProto(model.CreatedAt)
	updated, _ := ptypes.TimestampProto(model.UpdatedAt)
	if model.Type == cardpayment.PaymentType_UNKNOWN_PAYMENT_TYPE {
		model.Type = cardpayment.PaymentType_CARD
	}

	return &cardpayment.Receipt{
		Id:           model.Id,
		Type:         model.Type,
		ClientSecret: model.ClientSecret,
		UserId:       model.UserId,
		UserEmail:    model.UserEmail,
		Items:        items,
		Property:     model.Property,
		Status:       model.Status,
		CreatedAt:    created,
		UpdatedAt:    updated,
	}
}

func RenderItemModel(model *CardpaymentItemModel) *cardpayment.ReceiptItem {
	created, _ := ptypes.TimestampProto(model.CreatedAt)
	updated, _ := ptypes.TimestampProto(model.UpdatedAt)
	if model.Confirm == nil {
		model.Confirm = &ConfirmModel{}
	}
	if model.ChargeType == cardpayment.PaymentType_UNKNOWN_PAYMENT_TYPE {
		model.ChargeType = cardpayment.PaymentType_CARD
	}
	return &cardpayment.ReceiptItem{
		ChargeId:      model.ChargeId,
		ChargeType:    model.ChargeType,
		ClientSecret:  model.ClientSecret,
		UserId:        model.UserId,
		UserEmail:     model.UserEmail,
		Property:      model.Property,
		ChargeStatus:  model.ChargeStatus,
		CreatedAt:     created,
		UpdatedAt:     updated,
		InvoiceNumber: model.InvoiceNumber,

		ServiceId:   &cardpayment.ReceiptItem_ServiceIdValue{ServiceIdValue: model.ServiceId},
		Id:          &cardpayment.ReceiptItem_IdValue{IdValue: model.Id},
		Type:        &cardpayment.ReceiptItem_TypeValue{TypeValue: model.Type},
		Code:        &cardpayment.ReceiptItem_CodeValue{CodeValue: model.Code},
		Status:      &cardpayment.ReceiptItem_StatusValue{StatusValue: model.Status},
		Price:       &cardpayment.ReceiptItem_PriceValue{PriceValue: model.Price},
		Description: &cardpayment.ReceiptItem_DescriptionValue{DescriptionValue: model.Description},
		Primary:     &cardpayment.ReceiptItem_PrimaryValue{PrimaryValue: model.Primary},
		Confirm: &cardpayment.ReceiptItem_ConfirmValue{ConfirmValue: &extras.ServiceConfirm{
			Admin:       model.Confirm.Admin,
			Confirmed:   model.Confirm.Confirmed,
			ConfirmedAt: ToUnixTime(model.Confirm.ConfirmedAt),
		}},
	}
}

func RenderAndSendInvoice(userId string, receipt *CardpaymentModel) bool {
	var payment PaymentForInvoice
	var items []*extras.Service
	createdAt, _ := ptypes.TimestampProto(receipt.CreatedAt)

	for _, itm := range receipt.Items {
		items = append(items, RenderServiceModel(itm))
	}
	payment.Services = items
	payment.InvoiceNumber = receipt.InvoiceNumber
	payment.CreatedAt = createdAt

	user, err := AccountService.Get(context.Background(), &account.GetRequest{
		Id: userId,
	})
	if err != nil {
		fmt.Printf("get account for invoice sending err %v\n", err)
		return false
	}

	cs := struct {
		Now     string             `json:"now"`
		User    *account.Account   `json:"user"`
		Payment *PaymentForInvoice `json:"payment"`
	}{
		Now:     time.Now().Format(time.RFC1123),
		User:    user.Account,
		Payment: &payment,
	}

	bs, err := bind.BindToJson(cs, &bind.Config{
		GetOmmitempty: false,
		SetOmmitempty: false,
	})
	if err != nil {
		fmt.Printf("bind to json for invoice sending err %v\n", err)
		return false
	}
	resp, err := ContractService.RenderInvoice(context.Background(), &contract.RenderInvoiceRequest{
		Format:     contract.ContractFormat_PDF,
		ValuesJson: bs,
	})
	if err != nil {
		fmt.Printf("render for invoice sending err %v\n", err)
		return false
	}
	_, err = NotificationService.SendEvent(context.Background(), &notification.SendEventRequest{
		User:     userId,
		Event:    notification.EventType_EVENT_PAYMENT_MASHROOM_INVOICE,
		Channels: []notification.EventChannel{notification.EventChannel_EVENT_CHANNEL_EMAIL},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_FILE,
				EmbedValue: resp.Fileid,
			},
		},
	})
	if err != nil {
		fmt.Printf("sending invoice error %v\n", err)
		return false
	}

	return true
}

func (server *serverImpl) GetReceipt(ctx context.Context, request *cardpayment.GetReceiptRequest) (*cardpayment.GetReceiptResponse, error) {
	m, err := server.repo.GetByID(request.GetId())
	if err != nil {
		return nil, err
	}
	return &cardpayment.GetReceiptResponse{
		Receipt: RenderModel(m),
	}, nil
}

func (server *serverImpl) GetUser(ctx context.Context, request *cardpayment.GetUserRequest) (*cardpayment.GetUserResponse, error) {
	rcpts, err := server.repo.GetByUser(request.GetUser())
	if err != nil {
		return nil, err
	}
	resp := &cardpayment.GetUserResponse{}
	for _, r := range rcpts {
		resp.Payments = append(resp.Payments, RenderModel(r))
	}
	return resp, nil
}

func (server *serverImpl) GetProperty(ctx context.Context, request *cardpayment.GetPropertyRequest) (*cardpayment.GetPropertyResponse, error) {
	rcpts, err := server.repo.GetByUserProperty(request.GetUser(), request.GetProperty())
	if err != nil {
		return nil, err
	}
	resp := &cardpayment.GetPropertyResponse{}
	for _, r := range rcpts {
		resp.Payments = append(resp.Payments, RenderModel(r))
	}
	return resp, nil
}

func (server *serverImpl) ReceiptList(ctx context.Context, request *cardpayment.ReceiptListRequest) (*cardpayment.ReceiptListResponse, error) {
	var rcpts []*CardpaymentModel
	var err error
	user := request.GetUser()
	prop := request.GetProperty()
	if (user != "") && (prop == "") {
		rcpts, err = server.repo.GetByUser(user)
	} else if prop != "" {
		if user != "" {
			rcpts, err = server.repo.GetByUserProperty(user, prop)
		} else {
			rcpts, err = server.repo.GetByProperty(prop)
		}
	}
	if err != nil {
		return nil, err
	}
	resp := &cardpayment.ReceiptListResponse{}
	for _, r := range rcpts {
		resp.Payments = append(resp.Payments, RenderModel(r))
	}
	return resp, nil
}

func (server *serverImpl) ListFilter(ctx context.Context, req *cardpayment.ReceiptListFilterRequest) (*cardpayment.ReceiptListResponse, error) {
	fields := log.Fields{}
	log.WithFields(fields).Debug("list filter requested")
	params := req.GetParamsValue()

	rcpts, total, err := server.repo.FilterCharges(params, req.GetAdmin())
	if err != nil {
		log.WithFields(fields).WithError(err).Error("failed to list filtered models")
		return nil, err
	}
	log.WithFields(fields).Debug("list filter succeeded")

	resp := &cardpayment.ReceiptListResponse{
		Success: true,
		Total:   total,
	}
	for _, r := range rcpts {
		resp.Payments = append(resp.Payments, RenderModel(r))
	}
	return resp, nil
}

func (server *serverImpl) ListItemsFilter(ctx context.Context, req *cardpayment.ReceiptListFilterRequest) (*cardpayment.ReceiptItemsListResponse, error) {
	fields := log.Fields{}
	log.WithFields(fields).Debug("list filter requested")
	params := req.GetParamsValue()

	rcpts, total, nonConfirmed, err := server.repo.FilterChargeItems(params, req.GetAdmin())
	if err != nil {
		log.WithFields(fields).WithError(err).Error("failed to list filtered models")
		return nil, err
	}
	log.WithFields(fields).Debug("list filter succeeded")

	resp := &cardpayment.ReceiptItemsListResponse{
		Success: true,
		Total:   total,
	}
	nonConfirmed = 0
	for i, r := range rcpts {
		resp.Payments = append(resp.Payments, RenderItemModel(r))
		if !resp.Payments[i].GetConfirmValue().GetConfirmed() {
			nonConfirmed = nonConfirmed + 1
		}
	}
	resp.NonConfirmed = nonConfirmed

	return resp, nil
}

func (server *serverImpl) GetItem(ctx context.Context, req *cardpayment.ReceiptItemRequest) (*cardpayment.ReceiptItem, error) {
	fields := log.Fields{}
	log.WithFields(fields).Debug("charged item requested")

	item, err := server.repo.GetChargeItem(req.GetChargeId(), req.GetId())
	if err != nil {
		log.WithFields(fields).WithError(err).Error("failed to get a charged item")
		return nil, err
	}
	log.WithFields(fields).Debug("charged item found")
	return RenderItemModel(item), nil
}

func (server *serverImpl) ConfirmItem(ctx context.Context, req *cardpayment.ReceiptItemConfirmRequest) (*cardpayment.ReceiptItem, error) {
	fields := log.Fields{}
	log.WithFields(fields).Debug("charged item requested for update")
	item, err := server.repo.ConfirmChargeItem(req.GetChargeId(), req.GetId(), req.GetAdmin(), req.GetConfirm())
	if err != nil {
		log.WithFields(fields).WithError(err).Error("failed to get a charged item")
		return nil, err
	}
	log.WithFields(fields).Debug("charged item found")
	return RenderItemModel(item), nil
}

func (server *serverImpl) MarkReceiptPaid(ctx context.Context, request *cardpayment.MarkReceiptPaidRequest) (*cardpayment.MarkReceiptPaidResponse, error) {
	m, err := server.repo.GetByID(request.GetId())
	if err != nil {
		return nil, err
	}
	if m.UserId != request.GetUserId() {
		return nil, ErrNotOwned
	}
	if m.Status == cardpayment.PaymentStatus_CANCELLED {
		return nil, ErrCancelled
	}
	m.Status = cardpayment.PaymentStatus_SUCCEEDED

	createInvoice := true
	for _, item := range m.Items {
		if item.Role != extras.ServiceRole_LANDLORD_LISTING {
			createInvoice = false
			break
		}
	}

	if createInvoice {
		paymentType := contract.PaymentType_CARD
		if m.Type == cardpayment.PaymentType_BANK_ACCOUNT {
			paymentType = contract.PaymentType_BANK_ACCOUNT
		}
		resp, err := ContractService.CreateInvoice(ctx, &contract.CreateInvoiceRequest{
			Payment: m.Id,
			Type:    paymentType,
			UserId:  m.UserId,
		})

		if err != nil {
			return &cardpayment.MarkReceiptPaidResponse{Success: false}, err
		}

		m.InvoiceNumber = resp.Id

		renderSuccess := RenderAndSendInvoice(m.UserId, m)
		if !renderSuccess {
			fmt.Printf("RenderAndSendInvoice method error")
		}
	}

	err = server.repo.Update(m)
	if err != nil {
		return nil, err
	}
	server.IssueUpdate(m)

	landlordListing := true
	var descSlice []string
	var costFloat float64
	for _, item := range m.Items {
		descSlice = append(descSlice, item.Description)
		costFloat += item.Price
		if item.Role != extras.ServiceRole_LANDLORD_LISTING {
			landlordListing = false
		}
	}
	cost := strconv.FormatFloat(costFloat, 'f', 2, 64)

	_, err = NotificationService.SendEvent(context.Background(), &notification.SendEventRequest{
		User:     m.UserId,
		Event:    notification.EventType_EVENT_PAYMENT_PAID_EXTRA_LANDLORD,
		Channels: []notification.EventChannel{notification.EventChannel_EVENT_CHANNEL_PLATFORM, notification.EventChannel_EVENT_CHANNEL_SMS, notification.EventChannel_EVENT_CHANNEL_EMAIL},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PAYMENT_COST,
				EmbedValue: cost,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PAYMENT_ITEM,
				EmbedValue: strings.Join(descSlice, ", "),
			},
		},
	})

	if err != nil {
		log.WithError(err).Error("sending paid extra notification")
	}

	if landlordListing {
		respRender, err := PropertyService.RenderVibrant(context.Background(), &property.RenderVibrantRequest{Receipt: request.Id})
		if err != nil || !respRender.Success {
			log.WithError(err).Error("render vibrant error")
		}
	}

	return &cardpayment.MarkReceiptPaidResponse{Success: true}, nil
}

func (server *serverImpl) MarkReceiptPaidViaBank(ctx context.Context, request *cardpayment.MarkReceiptPaidRequest) (*cardpayment.MarkReceiptPaidResponse, error) {
	m, err := server.repo.GetByID(request.GetId())
	if err != nil {
		return nil, err
	}
	if m.UserId != request.GetUserId() {
		return nil, ErrNotOwned
	}
	if m.Status == cardpayment.PaymentStatus_CANCELLED {
		return nil, ErrCancelled
	}
	m.Type = cardpayment.PaymentType_BANK_ACCOUNT
	m.Status = cardpayment.PaymentStatus_SUCCEEDED
	err = server.repo.Update(m)
	if err != nil {
		return nil, err
	}
	server.IssueUpdate(m)
	return &cardpayment.MarkReceiptPaidResponse{Success: true}, nil
}

func (server *serverImpl) DownloadInvoicesCsv(ctx context.Context, request *cardpayment.DownloadInvoicesCsvRequest) (*cardpayment.DownloadInvoicesCsvResponse, error) {
	//var columns = []string{
	//	"*ContactName",
	//	"EmailAddress",
	//	"POAddressLine1",
	//	"POAddressLine2",
	//	"POAddressLine3",
	//	"POAddressLine4",
	//	"POCity",
	//	"PORegion",
	//	"POPostalCode",
	//	"POCountry",
	//	"*InvoiceNumber",
	//	"Reference",
	//	"*InvoiceDate",
	//	"*DueDate",
	//	"Total",
	//	"InventoryItemCode",
	//	"*Description",
	//	"*Quantity",
	//	"*UnitAmount",
	//	"*AccountCode",
	//	"*TaxType",
	//	"TaxAmount",
	//	"TrackingName1",
	//	"TrackingOption1",
	//	"TrackingName2",
	//	"TrackingOption2",
	//	"Currency",
	//	"BrandingTheme",
	//}
	var sb strings.Builder
	header := "*ContactName, EmailAddress, POAddressLine1, POAddressLine2, POAddressLine3, POAddressLine4, POCity," +
		" PORegion, POPostalCode, POCountry, *InvoiceNumber, Reference, *InvoiceDate, *DueDate," +
		" Total, InventoryItemCode, *Description, *Quantity, *UnitAmount, Discount, *AccountCode," +
		" *TaxType, TaxAmount, TrackingName1, TrackingOption1, TrackingName2, TrackingOption2, Currency, BrandingTheme\n"
	respListItems, err := server.ListItemsFilter(ctx, &cardpayment.ReceiptListFilterRequest{
		Admin: true,
	})
	if err != nil || !respListItems.GetSuccess() {
		return &cardpayment.DownloadInvoicesCsvResponse{
			Success: false,
		}, err
	}

	switch request.GetInvoiceType() {
	//Add-on products - containing the details of payments made for EPCs, Gas Safety Certificates, Inventory, Inventory & check-in bundle, check-out, photos+floorplan, Deposit Replacement Product
	case cardpayment.InvoiceType_ADD_ON_PRODUCTS:
		{
			if respListItems.GetPayments() != nil {
				//data := [][]string{}
				sb.WriteString(header)
				for _, itm := range respListItems.GetPayments() {
					if itm.GetTypeValue() == extras.ServiceType_PHOTOS_FLOOR_PLAN || itm.GetTypeValue() == extras.ServiceType_CERTIFICATE ||
						(itm.GetTypeValue() == extras.ServiceType_DEPOSIT && itm.GetCodeValue() == "tenant_deposit_replacement") ||
						(itm.GetTypeValue() == extras.ServiceType_BUNDLE && (itm.GetCodeValue() == "landlord_inventory_check_in_furnished" || itm.GetCodeValue() == "landlord_inventory_check_in_unfurnished")) ||
						(itm.GetTypeValue() == extras.ServiceType_INVENTORY_REPORT && (itm.GetCodeValue() == "landlord_inventory_unfurnished" ||
							itm.GetCodeValue() == "landlord_inventory_furnished" || itm.GetCodeValue() == "landlord_check_out")) {

						//row := make([]string, len(columns))
						respAcc, err := AccountService.Get(ctx, &account.GetRequest{Id: itm.GetUserId()})
						if err != nil {
							return nil, err
						}
						var amount float64
						m, _ := server.repo.GetByID(itm.GetChargeId())
						for _, reciept := range RenderModel(m).Items {
							amount += reciept.GetPriceValue()
						}
						contactName := fmt.Sprintf("%s %s", respAcc.GetAccount().GetDetailsValue().GetFirstNameValue(), respAcc.GetAccount().GetDetailsValue().GetLastNameValue())
						invoiceDate := time.Unix(itm.CreatedAt.Seconds, int64(itm.CreatedAt.Nanos))
						body := fmt.Sprintf("%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %d, %s, %s, %s, %.2f, %s, %s, %d, %.2f, %d, %d, %s, %.2f, %s, %s, %s, %s, %s, %s\n", contactName, itm.GetUserEmail(), "", "", "", "", "", "-", "",
							"UK", itm.GetInvoiceNumber(), itm.GetChargeId(), invoiceDate.Format("02/01/2006 15:04:05"), invoiceDate.Format("02/01/2006 15:04:05"), amount, itm.GetIdValue(), itm.GetDescriptionValue(),
							1, (itm.GetPriceValue()/6)*5, 0, 200, "20% (VAT on Income)", itm.GetPriceValue()/6, "Activity", "Sales", "", "", "GBP", "Standard")
						sb.WriteString(body)
					}
				}
			}
		}
	//Insurance products - containing the details of payments made for LEI & Rent Guarantee insurance, Property Emergency Insurance
	case cardpayment.InvoiceType_INSURANCE_PRODUCTS:
		{
			if respListItems.GetPayments() != nil {
				//data := [][]string{}
				sb.WriteString(header)
				for _, itm := range respListItems.GetPayments() {
					if itm.GetCodeValue() == "landlord_rent_guarantee" || itm.GetCodeValue() == "landlord_property_emergency_insurance" {

						//row := make([]string, len(columns))
						respAcc, err := AccountService.Get(ctx, &account.GetRequest{Id: itm.GetUserId()})
						if err != nil {
							return nil, err
						}
						var amount float64
						m, _ := server.repo.GetByID(itm.GetChargeId())
						for _, reciept := range RenderModel(m).Items {
							amount += reciept.GetPriceValue()
						}
						contactName := fmt.Sprintf("%s %s", respAcc.GetAccount().GetDetailsValue().GetFirstNameValue(), respAcc.GetAccount().GetDetailsValue().GetLastNameValue())
						invoiceDate := time.Unix(itm.CreatedAt.Seconds, int64(itm.CreatedAt.Nanos))
						body := fmt.Sprintf("%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %d, %s, %s, %s, %.2f, %s, %s, %d, %.2f, %d, %d, %s, %.2f, %s, %s, %s, %s, %s, %s\n", contactName, itm.GetUserEmail(), "", "", "", "", "", "-", "",
							"UK", itm.GetInvoiceNumber(), itm.GetChargeId(), invoiceDate.Format("02/01/2006 15:04:05"), invoiceDate.Format("02/01/2006 15:04:05"), amount, itm.GetIdValue(), itm.GetDescriptionValue(),
							1, (itm.GetPriceValue()/112)*100, 0, 200, "12% (VAT on Income)", (itm.GetPriceValue()/112)*12, "Activity", "Sales", "", "", "GBP", "Standard")
						sb.WriteString(body)
					}
				}
			}

		}
	case cardpayment.InvoiceType_UNKNOWN_INVOICE_TYPE:
		{
			return nil, errors.New("unknown invoice type for rendering")
		}
	default:
		return nil, errors.New("unknown invoice type for rendering")

	}

	bt := []byte(sb.String())

	upload, err := DocumentService.FileUpload(ctx)
	if err != nil {
		return nil, err
	}
	defer func() {
		_ = upload.CloseSend()
	}()

	now := time.Now()
	name := fmt.Sprintf("testaddon-%04d-%02d-%02d.%s", now.Year(), now.Month(), now.Day(), "csv")
	desc := fmt.Sprintf("addon %04d-%02d-%02d, rendered in %s", now.Year(), now.Month(), now.Day(), "csv")
	shahash := sha256.New()
	_, _ = shahash.Write(bt)
	err = upload.Send(&document.FileEvent{
		Init: &document.FileInit{
			File: &document.File{
				Name:        name,
				Mime:        "text/csv",
				Size:        int64(len(bt)),
				CreatedAt:   ToUnixTime(now),
				Sha256:      shahash.Sum(nil),
				Description: desc,
			},
		},
	})

	if err != nil {
		_ = upload.CloseSend()
		return nil, err
	}

	i := 0
	for i = 0; i < len(bt); i += 2048 {
		chunk := bt[i:]
		if len(chunk) > 2048 {
			chunk = chunk[:2048]
		}
		err = upload.Send(&document.FileEvent{
			Data: &document.FileData{
				Data: chunk,
			},
		})
		if err != nil {
			_ = upload.CloseSend()
			return nil, err
		}
	}

	resp, err := upload.CloseAndRecv()
	if err != nil {
		return nil, err
	}
	if !resp.Success {
		return nil, errors.New("upload failure")
	}

	return &cardpayment.DownloadInvoicesCsvResponse{
		File:    resp.GetFile().GetId(),
		Success: resp.GetSuccess(),
	}, nil
}

func (server *serverImpl) Background() {
	if err := server.repo.FixItemIds(); err == nil {
		// TODO: log?
	}
	// background processes
}

func (server *serverImpl) Serve(addr string) {
	if listener, err := net.Listen("tcp", addr); err != nil {
		panic(err)
	} else {
		grpcServer := grpc.NewServer("cardpayment")

		cardpayment.RegisterCardpaymentServiceServer(grpcServer, server)

		healthServer := version.NewHealthServer()
		health.RegisterHealthServiceServer(grpcServer, healthServer)

		log.Infoln("cardpayment started")
		if err := grpcServer.Serve(listener); err != nil {
			log.Errorln("gRPC error", err)
		}
	}
}

func NewServer(repo *Repo) Server {
	server := &serverImpl{repo: repo}
	server.ZBClient = ServeBpms(server)
	go server.Background()
	return server
}
